package de.slpnetwork.database;

import java.sql.*;
import java.util.Iterator;
import java.util.Map;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import de.slpnetwork.MCPassiveMode;
import de.slpnetwork.interfaces.IDataHandler;
import de.slpnetwork.utils.ArrayHelper;

import javax.annotation.Nullable;

public class SQLite implements IDataHandler {

    private Connection con;
    private MCPassiveMode plugin;

    public SQLite(MCPassiveMode plugin, String path) throws Exception {

        this.plugin = plugin;

        // setup connection
        try {
            this.con = connect(path);
        } catch (Exception ex) {
            throw new Exception("SQLite initailzatio failed due to exception" + ex.getLocalizedMessage(), ex.getCause());
        }
    }

    private Connection connect(String path) throws Exception {
        try {
            return DriverManager.getConnection("jdbc:sqlite:" + path);
        } catch (Exception ex) {
            throw new Exception("SQLite connection failed due to exception" + ex.getLocalizedMessage(), ex.getCause());
        }
    }

    /**
     * creates a table in the set database
     *
     * @param name the tables name
     * @throws SQLException
     */
    @Override
    public void createTable(String name, Map<String, String> fieldMap) {
        String fields = "";

        fields += "(";
        for (Map.Entry<String, String> entry : fieldMap.entrySet()) {
            System.out.println(entry.getKey() + ":" + entry.getValue());
            fields += entry.getKey() + " " + entry.getValue();
        }
        fields += ")";

        try {
            this.con.createStatement().executeQuery("CREATE TABLE IF NOT EXISTS " + name + " " + fields);
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }

    /**
     * executes a given query and returns a result of  type int
     *
     * @param stmt the query to execute
     * @param target    secondary argument used to define the required field
     * @return the found result of type int
     */
    @Override
    public int getInt(String stmt, String target) {
        try {
            int returning = 0;
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                returning = res.getInt(target);
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return 0;
    }

    @Override
    public int[] getInts(String stmt, String target) {
        try {
            int[] returning = {};
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                ArrayHelper.push(returning, res.getInt(target));
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public double getDouble(String stmt, @Nullable String target) {
        try {
            double returning = 0;
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                returning = res.getDouble(target);
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return 0;
    }

    @Override
    public double[] getDoubles(String stmt, String target) {
        try {
            double[] returning = {};
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                ArrayHelper.push(returning, res.getDouble(target));
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public String getString(String stmt, String target) {
        try {
            String returning = "";
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                returning = res.getString(target);
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return "";
    }

    @Override
    public String[] getStrings(String stmt, String target) {
        try {
            String[] returning = {};
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                ArrayHelper.push(returning, res.getString(target));
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * execute a given statement on a mysql database
     * @param stmt the statement used to query the database
     * @param target not used
     * @return if found values from the database of type boolean
     */
    @Override
    public boolean getBool(String stmt, String target) {
        try {
            boolean returning = false;
            ResultSet res = this.con.createStatement().executeQuery(stmt);

            while (res.next()) {
                returning = res.getBoolean(target);
            }

            return returning;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return false;
    }

    /**
     * execute a given statement on a mysql database
     * @param stmt the statement used to query the database
     * @param target not used
     * @return if found values from the database of type boolean as array
     */
    @Override
    public boolean[] getBools(String stmt, String target) {
        boolean[] result = {};

        try {
            ResultSet rs = this.con.createStatement().executeQuery(stmt);
            while (rs.next()) {
                result = ArrayHelper.push(result,rs.getBoolean(target));
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return result;
    }

    /**
     * executes a query and returns the unproccessed resultset
     * @param stmt the query given as a string
     * @return a resultset to continue proccessing with
     */
    @Override
    public ResultSet getResults(String stmt) throws Exception {
        try {
            return this.con.createStatement().executeQuery(stmt);
        } catch (SQLException e) {
            e.printStackTrace();
        }

        throw new Exception("reached end of function.. this should never happen");
    }


    @Override
    public JsonObject getResultAsJSON(String stmt) throws Exception {
        try {
            ResultSet rs = this.con.createStatement().executeQuery(stmt);
            JsonObject jsonObject = new JsonObject();

            while (rs.next()) {
                int totalColumns = rs.getMetaData().getColumnCount();

                for (int i = 1; i <= totalColumns; i++) {
                    jsonObject.add(rs.getMetaData().getColumnLabel(i), (JsonElement) rs.getObject(i));
                }
            }

        } catch (SQLException e) {
            e.printStackTrace();
        }

        throw new Exception("reached end of function.. this should never happen");
    }


    @Override
    public JsonObject[] getResultsAsJSON(String stmt) throws Exception {
        try {
            ResultSet rs = this.con.createStatement().executeQuery(stmt);
            JsonObject[] jsonArray = new JsonObject[0];

            while (rs.next()) {
                JsonObject jsonObject = new JsonObject();
                int totalColumns = rs.getMetaData().getColumnCount();

                for (int i = 1; i <= totalColumns; i++) {
                    jsonObject.add(rs.getMetaData().getColumnLabel(i), (JsonElement) rs.getObject(i));
                }

                ArrayHelper.push(jsonArray, jsonObject);
            }

        } catch (SQLException e) {
            e.printStackTrace();
        }

        throw new Exception("reached end of function.. this should never happen");
    }


    @Override
    public Map<Byte[], Byte[]> getResultsAsMap(String a1) {
        // TODO Auto-generated method stub
        throw new UnsupportedOperationException("Unimplemented method 'getResultsAsMap'");
    }

    @Override
    public boolean writeData(String tableName, Map<String, Byte[]> values) throws Exception {
        if (values.isEmpty()) {
            System.err.println("Data map is empty. No columns to insert values into.");
            return false;
        }

        if (tableName == "" ) {
            throw new Exception("found empty tableName");
        }

        StringBuilder labels = new StringBuilder();
        StringBuilder placeholders = new StringBuilder();

        Iterator<Map.Entry<String, Byte[]>> iterator = values.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, Byte[]> entry = iterator.next();
            labels.append(entry.getKey());
            placeholders.append("?");

            if (iterator.hasNext()) {
                labels.append(",");
                placeholders.append(",");
            }
        }

        String sql = "INSERT INTO " + tableName + " (" + labels + ") VALUES (" + placeholders + ")";

        try (PreparedStatement statement = this.con.prepareStatement(sql)) {
            int index = 1;
            for (Byte[] value : values.values()) {
                statement.setObject(index++, value);
            }

            int rowsInserted = statement.executeUpdate();
            return rowsInserted > 0;
        } catch (SQLException e) {
            e.printStackTrace();
            return false;
        }
    }

    @Override
    public boolean writeData(String tableName, String condition, Map<String, Byte[]> values) throws Exception {
        if (values.isEmpty()) {
            System.err.println("Data map is empty. No columns to insert values into.");
            return false;
        }

        if (tableName == "" ) {
            throw new Exception("found empty tableName");
        }

        if (condition == "" ) {
            throw new Exception("found empty condition");
        }

        StringBuilder labels = new StringBuilder();
        StringBuilder placeholders = new StringBuilder();

        Iterator<Map.Entry<String, Byte[]>> iterator = values.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, Byte[]> entry = iterator.next();
            labels.append(entry.getKey());
            placeholders.append("?");

            if (iterator.hasNext()) {
                labels.append(",");
                placeholders.append(",");
            }
        }

        String sql = "INSERT INTO " + tableName + " (" + labels + ") VALUES (" + placeholders + ") " + condition;

        try (PreparedStatement statement = this.con.prepareStatement(sql)) {
            int index = 1;
            for (Byte[] value : values.values()) {
                statement.setObject(index++, value);
            }

            int rowsInserted = statement.executeUpdate();
            return rowsInserted > 0;
        } catch (SQLException e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * counts elements of a given resultset
     * @param resultSet the requested resultset to count
     * @return the counted size of a resultset
     * @throws SQLException
     */
    private int countResult(ResultSet resultSet) throws SQLException {
        int size = 0;

        while(resultSet.next()) {
            size++;
        }

        return size;
    }
}
